
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"

extern void	Error(/* char *msg */);

#define BITS	(sizeof(Cell)*8)

static int	GS2BW(Nx,Ny,GS,lower,upper,BW,X1,Y1,X2,Y2)
int		Nx,Ny,lower,upper,*X1,*Y1,*X2,*Y2;
Cell		**GS,**BW;
{
	int	y,x;
	Cell	*ptr,bit,off;

	*X1=Nx; *Y1=Ny; *X2=(*Y2)=(-1);
	for (y=0; y<Ny; y++) {
	    ptr=BW[y/BITS]; bit=(Cell)(1<<(y%BITS)); off=(Cell)(~bit);
	    for (x=0; x<Nx; x++)
		if (GS[y][x]>=lower && GS[y][x]<=upper) {
		    ptr[x]|=bit; if (x<*X1) *X1=x; if (x>*X2) *X2=x;
				 if (y<*Y1) *Y1=y; if (y>*Y2) *Y2=y;
		}
		else
		    ptr[x]&=off;
	}
	return *X1<=*X2 && *Y1<=*Y2;
}

typedef unsigned int	Card;

static Card	Search(table,label)
Card		*table,label;
{
	while (table[label]) label=table[label];

	return label;
}

#define MA(noe,val)	malloc((size_t)(noe)*sizeof(val))

static void	CountCores(X1,Y1,X2,Y2,BW,last,title)
int		X1,Y1,X2,Y2;
Cell		**BW;
Card		*last;
char		*title;
{
	Card	layer,labels,*table,pixels,label,cores;
	int	x1,y1,x2,y2,x,y;
	Cell	*ptr,bit,*ptr2,bit2,*ptr1,bit1,off;

	for (layer=0;; layer++) {
	    x1=X1; y1=Y1; x2=X2; y2=Y2; last[x1-1]=0;

	    labels=0;
	    for (x=x1; x<=x2; x++) last[x]=0;
	    for (y=y1; y<=y2; y++) {
		ptr=BW[y/BITS]; bit=(Cell)(1<<(y%BITS));
		for (x=x1; x<=x2; x++)
		    if (ptr[x]&bit) {
			if (!last[x] && !last[x-1]) ++labels;

			last[x]=1;
		    }
		    else
			last[x]=0;
	    }
	    if ((table=(Card *)MA(labels,*table))==NULL)
		Error("no memory for 'table'.");

	    --table;

	    pixels=0; label=1;
	    for (x=x1; x<=x2; x++) last[x]=0;
	    for (y=y1; y<=y2; y++) {
		ptr=BW[y/BITS]; bit=(Cell)(1<<(y%BITS));
		for (x=x1; x<=x2; x++)
		    if (ptr[x]&bit) {
			++pixels;

			if (last[x]) {
			    last[x]=Search(table,last[x]);
			    if (last[x-1] && last[x-1]!=last[x])
				last[x-1]=table[last[x-1]]=last[x];
			}
			else if (last[x-1])
			    last[x]=last[x-1];
			else
			    table[last[x]=label++]=0;
		    }
		    else
			last[x]=0;
	    }
	    cores=0; while (--label>=1) if (!table[label]) ++cores;

	    free(table+1);

	    printf("%s%lu\t%lu\t%lu\n",title,(unsigned long)layer,
					     (unsigned long)pixels,
					     (unsigned long)cores);

	    X1=x2+1; Y1=y2+1; X2=x1-1; Y2=y1-1;
	    ptr2=BW[y1/BITS]; bit2=(Cell)(1<<(y1%BITS));
	    for (x=x1; x<=x2; x++) last[x]=0;
	    for (y=y1; y<=y2; y++) {
		ptr1=ptr2; bit1=bit2; off=(Cell)(~bit1);
		ptr2=BW[(y+1)/BITS]; bit2=(Cell)(1<<((y+1)%BITS));
		for (x=x1; x<=x2; x++)
		    if (ptr1[x]&bit1) {
			if (!last[x] ||
			    !last[x-1] ||
			    y==y2 || !(ptr2[x]&bit2) ||
			    x==x2 || !(ptr1[x+1]&bit1))
			    ptr1[x]&=off;
			else {
			    if (x<X1) X1=x; if (x>X2) X2=x;
			    if (y<Y1) Y1=y; if (y>Y2) Y2=y;
			}
			last[x]=1;
		    }
		    else
			last[x]=0;
	    }
	    if (X1>X2 || Y1>Y2) break;
	}
}

#define TITLE_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,X1,Y1,X2,Y2,lower,upper,y;
	Cell	**GS,**BW;
	Card	*last;
	char	title[TITLE_LEN],*nmfbw="no memory for 'BW'.";

	if (argc!=2 && argc!=4) Error("usage : cc_2d TIFF {lower upper}");

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&GS,NULL);

	if ((last=(Card *)MA(Nx+1,*last))==NULL) Error("no memory for 'last'.");

	--last;

	if (argc==4) {
	    if (!GS2BW(Nx,Ny,GS,atoi(argv[2]),
				atoi(argv[3]),GS,&X1,&Y1,&X2,&Y2))
		Error("no object pixel.");

	    CountCores(X1,Y1,X2,Y2,GS,last,"");
	}
	else {
	    if ((BW=(Cell **)MA(Ny/BITS+1,*BW))==NULL) Error(nmfbw);

	    for (y=0; y<Ny; y+=BITS)
		if ((BW[y/BITS]=(Cell *)MA(Nx,**BW))==NULL) Error(nmfbw);

	    while (scanf("%d %d",&lower,&upper)==2)
		if (GS2BW(Nx,Ny,GS,lower,upper,BW,&X1,&Y1,&X2,&Y2)) {
		    sprintf(title,"%d\t%d\t",lower,upper);
		    CountCores(X1,Y1,X2,Y2,BW,last,title);
		}
	}
	return 0;
}
